@@ -1,7 +1,7 @@
 #ifndef PROJECT_TREE_BUILDER__MSVC_PRJ_UTILS__HPP
 #define PROJECT_TREE_BUILDER__MSVC_PRJ_UTILS__HPP
 
-/* $Id: msvc_prj_utils.hpp 407291 2013-07-19 14:02:44Z gouriano $
+/* $Id: msvc_prj_utils.hpp 430659 2014-03-27 18:13:30Z gouriano $
  * ===========================================================================
  *
  *                            PUBLIC DOMAIN NOTICE
@@ -63,6 +63,7 @@ public:
 
     CProjKey(void);
     CProjKey(TProjType type, const string& project_id);
+    CProjKey(TProjType type, const string& project_id, const string& id_suffix);
     CProjKey(const CProjKey& key);
     CProjKey& operator= (const CProjKey& key);
     ~CProjKey(void);
@@ -73,10 +74,12 @@ public:
 
     TProjType     Type(void) const;
     const string& Id  (void) const;
+    string FullId  (void) const;
 
 private:
     TProjType m_Type;
     string    m_Id;
+    string    m_Suffix;
 
 };
 
@@ -251,7 +254,8 @@ public:
         eMsvc800,
         eMsvc900,
         eMsvc1000,
-        eMsvc1100,
+        eMsvc1100,      // vs2012
+        eMsvc1200,      // vs2013
         eXCode30,
         eMsvcNone
     };
@@ -610,6 +614,11 @@ private:
 };
 
 
+class CProjItem;
+class CDataToolGeneratedSrc;
+const CDataToolGeneratedSrc*
+IsProducedByDatatool(const string&  src_path_abs, const CProjItem& project);
+
 END_NCBI_SCOPE
 
 #endif //PROJECT_TREE_BUILDER__MSVC_PRJ_UTILS__HPP
