@@ -1,4 +1,4 @@
-/*  $Id: cgiapp.cpp 190528 2010-05-03 00:56:06Z kazimird $
+/*  $Id: cgiapp.cpp 448586 2014-10-07 16:53:03Z ivanov $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
@@ -46,6 +46,8 @@
 #include <cgi/cgi_exception.hpp>
 #include <cgi/cgi_serial.hpp>
 #include <cgi/error_codes.hpp>
+#include <signal.h>
+
 
 #ifdef NCBI_OS_UNIX
 #  include <unistd.h>
@@ -61,19 +63,43 @@ BEGIN_NCBI_SCOPE
 NCBI_PARAM_DECL(bool, CGI, Print_Http_Referer);
 NCBI_PARAM_DEF_EX(bool, CGI, Print_Http_Referer, true, eParam_NoThread,
                   CGI_PRINT_HTTP_REFERER);
-static NCBI_PARAM_TYPE(CGI, Print_Http_Referer) s_PrintRefererParam;
+typedef NCBI_PARAM_TYPE(CGI, Print_Http_Referer) TPrintRefererParam;
 
 
 NCBI_PARAM_DECL(bool, CGI, Print_User_Agent);
 NCBI_PARAM_DEF_EX(bool, CGI, Print_User_Agent, true, eParam_NoThread,
                   CGI_PRINT_USER_AGENT);
-static NCBI_PARAM_TYPE(CGI, Print_User_Agent) s_PrintUserAgentParam;
+typedef NCBI_PARAM_TYPE(CGI, Print_User_Agent) TPrintUserAgentParam;
 
 
 NCBI_PARAM_DECL(bool, CGI, Print_Self_Url);
 NCBI_PARAM_DEF_EX(bool, CGI, Print_Self_Url, true, eParam_NoThread,
                   CGI_PRINT_SELF_URL);
-static NCBI_PARAM_TYPE(CGI, Print_Self_Url) s_PrintSelfUrlParam;
+typedef NCBI_PARAM_TYPE(CGI, Print_Self_Url) TPrintSelfUrlParam;
+
+NCBI_PARAM_DECL(bool, CGI, Allow_Sigpipe);
+NCBI_PARAM_DEF_EX(bool, CGI, Allow_Sigpipe, false, eParam_NoThread,
+                  CGI_ALLOW_SIGPIPE);
+typedef NCBI_PARAM_TYPE(CGI, Allow_Sigpipe) TParamAllowSigpipe;
+
+// Log any broken connection with status 299 rather than 499.
+NCBI_PARAM_DEF_EX(bool, CGI, Client_Connection_Interruption_Okay, false,
+                  eParam_NoThread,
+                  CGI_CLIENT_CONNECTION_INTERRUPTION_OKAY);
+
+// Severity for logging broken connection errors.
+NCBI_PARAM_ENUM_ARRAY(EDiagSev, CGI, Client_Connection_Interruption_Severity)
+{
+    {"Info", eDiag_Info},
+    {"Warning", eDiag_Warning},
+    {"Error", eDiag_Error},
+    {"Critical", eDiag_Critical},
+    {"Fatal", eDiag_Fatal},
+    {"Trace", eDiag_Trace}
+};
+NCBI_PARAM_ENUM_DEF_EX(EDiagSev, CGI, Client_Connection_Interruption_Severity,
+                       eDiag_Critical, eParam_NoThread,
+                       CGI_CLIENT_CONNECTION_INTERRUPTION_SEVERITY);
 
 
 ///////////////////////////////////////////////////////
@@ -101,19 +127,16 @@ ERW_Result CCGIStreamReader::Read(void*   buf,
                                   size_t  count,
                                   size_t* bytes_read)
 {
-    streamsize x_read = CStreamUtils::Readsome(m_IStr, (char*)buf, count);
+    size_t x_read = (size_t)CStreamUtils::Readsome(m_IStr, (char*)buf, count);
     ERW_Result result;
-    if (x_read < 0) {
-        result = eRW_Error;
-    }
-    else if (x_read > 0) {
+    if (x_read > 0  ||  count == 0) {
         result = eRW_Success;
     }
     else {
-        result = eRW_Eof;
+        result = m_IStr.eof() ? eRW_Eof : eRW_Error;
     }
     if (bytes_read) {
-        *bytes_read = x_read < 0 ? 0 : x_read;
+        *bytes_read = x_read;
     }
     return result;
 }
@@ -165,6 +188,11 @@ CCgiApplication* CCgiApplication::Instance(void)
 }
 
 
+extern "C" void SigTermHandler(int)
+{
+}
+
+
 int CCgiApplication::Run(void)
 {
     // Value to return from this method Run()
@@ -180,8 +208,21 @@ int CCgiApplication::Run(void)
     // Make sure to restore old diagnostic state after the Run()
     CDiagRestorer diag_restorer;
 
-    // Compose diagnostics prefix
 #if defined(NCBI_OS_UNIX)
+    // Disable SIGPIPE if not allowed.
+    if ( !TParamAllowSigpipe::GetDefault() ) {
+        signal(SIGPIPE, SIG_IGN);
+        struct sigaction sigterm,  sigtermold;
+        memset(&sigterm, 0, sizeof(sigterm));
+        sigterm.sa_handler = SigTermHandler;
+        sigterm.sa_flags = SA_RESETHAND;
+        if (sigaction(SIGTERM, &sigterm, &sigtermold) == 0
+            &&  sigtermold.sa_handler != SIG_DFL) {
+            sigaction(SIGTERM, &sigtermold, 0);
+        }
+    }
+
+    // Compose diagnostics prefix
     PushDiagPostPrefix(NStr::IntToString(getpid()).c_str());
 #endif
     PushDiagPostPrefix(GetEnvironment().Get(m_DiagPrefixEnv).c_str());
@@ -199,12 +240,15 @@ int CCgiApplication::Run(void)
     //int orig_fd = -1;
     CNcbiStrstream result_copy;
     auto_ptr<CNcbiOstream> new_stream;
-    
+
     try {
         _TRACE("(CGI) CCgiApplication::Run: calling ProcessRequest");
         GetDiagContext().SetAppState(eDiagAppState_RequestBegin);
 
         m_Context.reset( CreateContext() );
+        _ASSERT(m_Context.get());
+        m_Context->CheckStatus();
+
         ConfigureDiagnostics(*m_Context);
         x_AddLBCookie();
         try {
@@ -237,10 +281,14 @@ int CCgiApplication::Run(void)
                     m_Context->GetResponse().SetOutput(new_stream.get());
                 }
                 GetDiagContext().SetAppState(eDiagAppState_Request);
-                result = ProcessRequest(*m_Context);
+                result = CCgiContext::ProcessCORSRequest(
+                    m_Context->GetRequest(), m_Context->GetResponse()) ?
+                    0 : ProcessRequest(*m_Context);
                 GetDiagContext().SetAppState(eDiagAppState_RequestEnd);
+                m_Context->GetResponse().Finalize();
                 if (result != 0) {
                     SetHTTPStatus(500);
+                    m_ErrorStatus = true;
                 } else {
                     if (m_Cache.get()) {
                         m_Context->GetResponse().Flush();
@@ -260,17 +308,33 @@ int CCgiApplication::Run(void)
             }
         }
         catch (CCgiException& e) {
-            if ( e.GetStatusCode() <  CCgiException::e200_Ok  ||
-                 e.GetStatusCode() >= CCgiException::e400_BadRequest ) {
-                throw;
+            if ( x_DoneHeadRequest() ) {
+                // Ignore errors after HEAD request has been finished.
+                GetDiagContext().SetAppState(eDiagAppState_RequestEnd);
+            }
+            else {
+                if ( e.GetStatusCode() <  CCgiException::e200_Ok  ||
+                     e.GetStatusCode() >= CCgiException::e400_BadRequest ) {
+                    throw;
+                }
+                GetDiagContext().SetAppState(eDiagAppState_RequestEnd);
+                // If for some reason exception with status 2xx was thrown,
+                // set the result to 0, update HTTP status and continue.
+                m_Context->GetResponse().SetStatus(e.GetStatusCode(),
+                                                   e.GetStatusMessage());
             }
-            GetDiagContext().SetAppState(eDiagAppState_RequestEnd);
-            // If for some reason exception with status 2xx was thrown,
-            // set the result to 0, update HTTP status and continue.
-            m_Context->GetResponse().SetStatus(e.GetStatusCode(),
-                                               e.GetStatusMessage());
             result = 0;
         }
+
+#ifdef NCBI_OS_MSWIN
+        // Logging - on MSWin this must be done before flushing the output.
+        if ( is_stat_log  &&  !skip_stat_log ) {
+            stat->Reset(start_time, result);
+            stat->Submit(stat->Compose());
+        }
+        is_stat_log = false;
+#endif
+
         _TRACE("CCgiApplication::Run: flushing");
         m_Context->GetResponse().Flush();
         _TRACE("CCgiApplication::Run: return " << result);
@@ -279,39 +343,58 @@ int CCgiApplication::Run(void)
     }
     catch (exception& e) {
         GetDiagContext().SetAppState(eDiagAppState_RequestEnd);
-        // Call the exception handler and set the CGI exit code
-        result = OnException(e, NcbiCout);
-        x_OnEvent(eException, result);
+        if ( x_DoneHeadRequest() ) {
+            // Ignore errors after HEAD request has been finished.
+            result = 0;
+            x_OnEvent(eSuccess, result);
+        }
+        else {
+            // Call the exception handler and set the CGI exit code
+            result = OnException(e, NcbiCout);
+            x_OnEvent(eException, result);
 
-        // Logging
-        {{
-            string msg = "(CGI) CCgiApplication::ProcessRequest() failed: ";
-            msg += e.what();
+            // Logging
+            {{
+                string msg = "(CGI) CCgiApplication::ProcessRequest() failed: ";
+                msg += e.what();
 
-            if ( is_stat_log ) {
-                stat->Reset(start_time, result, &e);
-                msg = stat->Compose();
-                stat->Submit(msg);
-                skip_stat_log = true; // Don't print the same message again
+                if ( is_stat_log ) {
+                    stat->Reset(start_time, result, &e);
+                    msg = stat->Compose();
+                    stat->Submit(msg);
+                    skip_stat_log = true; // Don't print the same message again
+                }
+            }}
+
+            // Exception reporting. Use different severity for broken connection.
+            ios_base::failure* fex = dynamic_cast<ios_base::failure*>(&e);
+            CNcbiOstream* os = m_Context.get() ? m_Context->GetResponse().GetOutput() : NULL;
+            if ((fex  &&  os  &&  !os->good())  ||  m_OutputBroken) {
+                if ( !TClientConnIntOk::GetDefault() ) {
+                    ERR_POST_X(13, Severity(TClientConnIntSeverity::GetDefault()) <<
+                        "Connection interrupted");
+                }
             }
-        }}
-
-        // Exception reporting
-        NCBI_REPORT_EXCEPTION_X(13, "(CGI) CCgiApplication::Run", e);
+            else {
+                NCBI_REPORT_EXCEPTION_X(13, "(CGI) CCgiApplication::Run", e);
+            }
+        }
     }
 
+#ifndef NCBI_OS_MSWIN
     // Logging
     if ( is_stat_log  &&  !skip_stat_log ) {
         stat->Reset(start_time, result);
-        string msg = stat->Compose();
-        stat->Submit(msg);
+        stat->Submit(stat->Compose());
     }
+#endif
 
     x_OnEvent(eEndRequest, 120);
     x_OnEvent(eExit, result);
 
-    if (orig_stream) 
+    if (m_Context.get()) {
         m_Context->GetResponse().SetOutput(NULL);
+    }
     return result;
 }
 
@@ -338,36 +421,46 @@ void CCgiApplication::LogRequest(void) const
 {
     const CCgiContext& ctx = GetContext();
     string str;
-    if ( s_PrintSelfUrlParam.Get() ) {
+    if ( TPrintSelfUrlParam::GetDefault() ) {
         // Print script URL
-        str = ctx.GetSelfURL();
-        if ( !str.empty() ) {
+        string self_url = ctx.GetSelfURL();
+        if ( !self_url.empty() ) {
             string args =
-                ctx.GetRequest().GetProperty(eCgi_QueryString);
-            if ( !args.empty() ) {
-                str += "?" + args;
+                ctx.GetRequest().GetRandomProperty("REDIRECT_QUERY_STRING", false);
+            if ( args.empty() ) {
+                args = ctx.GetRequest().GetProperty(eCgi_QueryString);
             }
-            GetDiagContext().Extra().
-                //SetType(kExtraType_CGI).
-                Print("SELF_URL", str);
+            if ( !args.empty() ) {
+                self_url += "?" + args;
+            }
+        }
+        // Add target url
+        string target_url = ctx.GetRequest().GetProperty(eCgi_ScriptName);
+        if ( !target_url.empty() ) {
+            string host = "http://" + GetDiagContext().GetHost();
+            string port = ctx.GetRequest().GetProperty(eCgi_ServerPort);
+            if (!port.empty()  &&  port != "80") {
+                host += ":" + port;
+            }
+            target_url = host + target_url;
+        }
+        if ( !self_url.empty()  ||  !target_url.empty() ) {
+            GetDiagContext().Extra().Print("SELF_URL", self_url).
+                Print("TARGET_URL", target_url);
         }
     }
     // Print HTTP_REFERER
-    if ( s_PrintRefererParam.Get() ) {
+    if ( TPrintRefererParam::GetDefault() ) {
         str = ctx.GetRequest().GetProperty(eCgi_HttpReferer);
         if ( !str.empty() ) {
-            GetDiagContext().Extra().
-                //SetType(kExtraType_CGI).
-                Print("HTTP_REFERER", str);
+            GetDiagContext().Extra().Print("HTTP_REFERER", str);
         }
     }
     // Print USER_AGENT
-    if ( s_PrintUserAgentParam.Get() ) {
+    if ( TPrintUserAgentParam::GetDefault() ) {
         str = ctx.GetRequest().GetProperty(eCgi_HttpUserAgent);
         if ( !str.empty() ) {
-            GetDiagContext().Extra().
-                //SetType(kExtraType_CGI).
-                Print("USER_AGENT", str);
+            GetDiagContext().Extra().Print("USER_AGENT", str);
         }
     }
 }
@@ -404,12 +497,12 @@ CNcbiResource& CCgiApplication::x_GetResource( void ) const
 NCBI_PARAM_DECL(bool, CGI, Merge_Log_Lines);
 NCBI_PARAM_DEF_EX(bool, CGI, Merge_Log_Lines, true, eParam_NoThread,
                   CGI_MERGE_LOG_LINES);
-static NCBI_PARAM_TYPE(CGI, Merge_Log_Lines) s_MergeLogLines;
+typedef NCBI_PARAM_TYPE(CGI, Merge_Log_Lines) TMergeLogLines;
 
 
 void CCgiApplication::Init(void)
 {
-    if ( s_MergeLogLines.Get() ) {
+    if ( TMergeLogLines::GetDefault() ) {
         // Convert multi-line diagnostic messages into one-line ones by default.
         SetDiagPostFlag(eDPF_PreMergeLines);
         SetDiagPostFlag(eDPF_MergeLines);
@@ -443,7 +536,7 @@ CCgiServerContext* CCgiApplication::LoadServerContext(CCgiContext& /*context*/)
 
 
 NCBI_PARAM_DECL(bool, CGI, Count_Transfered);
-NCBI_PARAM_DEF_EX(bool, CGI, Count_Transfered, false, eParam_NoThread,
+NCBI_PARAM_DEF_EX(bool, CGI, Count_Transfered, true, eParam_NoThread,
                   CGI_COUNT_TRANSFERED);
 typedef NCBI_PARAM_TYPE(CGI, Count_Transfered) TCGI_Count_Transfered;
 
@@ -469,6 +562,8 @@ CCgiContext* CCgiApplication::CreateContextWithFlags
  int               ofd,
  int               flags)
 {
+    m_OutputBroken = false; // reset failure flag
+
     int errbuf_size =
         GetConfig().GetInt("CGI", "RequestErrBufSize", 256, 0,
                            CNcbiRegistry::eReturn);
@@ -477,18 +572,20 @@ CCgiContext* CCgiApplication::CreateContextWithFlags
         if ( !inp ) {
             if ( !m_InputStream.get() ) {
                 m_InputStream.reset(
-                    new CRStream(new CCGIStreamReader(std::cin),
+                    new CRStream(new CCGIStreamReader(std::cin), 0, 0,
                                 CRWStreambuf::fOwnReader));
             }
             inp = m_InputStream.get();
+            ifd = 0;
         }
         if ( !out ) {
             if ( !m_OutputStream.get() ) {
                 m_OutputStream.reset(
-                    new CWStream(new CCGIStreamWriter(std::cout),
+                    new CWStream(new CCGIStreamWriter(std::cout), 0, 0,
                                 CRWStreambuf::fOwnWriter));
             }
             out = m_OutputStream.get();
+            ofd = 1;
             if ( m_InputStream.get() ) {
                 // If both streams are created by the application, tie them.
                 inp->tie(out);
@@ -546,12 +643,12 @@ CCgiApplication::CCgiApplication(void)
    m_HostIP(0), 
    m_Iteration(0),
    m_ArgContextSync(false),
+   m_OutputBroken(false),
    m_IsResultReady(true),
    m_ShouldExit(false),
-   m_RequestStartPrinted(false)
+   m_RequestStartPrinted(false),
+   m_ErrorStatus(false)
 {
-    // CGI applications should use /log for logging by default
-    CDiagContext::SetUseRootLog();
     // Disable system popup messages
     SuppressSystemMessageBox();
 
@@ -563,6 +660,7 @@ CCgiApplication::CCgiApplication(void)
     DisableArgDescriptions();
     RegisterDiagFactory("stderr", new CStderrDiagFactory);
     RegisterDiagFactory("asbody", new CAsBodyDiagFactory(this));
+    cerr.tie(0);
 }
 
 
@@ -581,7 +679,12 @@ int CCgiApplication::OnException(exception& e, CNcbiOstream& os)
     // Discriminate between different types of error
     string status_str = "500 Server Error";
     string message = "";
+
+    // Save current HTTP status. Later it may be changed to 299 or 499
+    // depending on this value.
+    m_ErrorStatus = CDiagContext::GetRequestContext().GetRequestStatus() >= 400;
     SetHTTPStatus(500);
+
     CException* ce = dynamic_cast<CException*> (&e);
     if ( ce ) {
         message = ce->GetMsg();
@@ -607,6 +710,11 @@ int CCgiApplication::OnException(exception& e, CNcbiOstream& os)
         message = e.what();
     }
 
+    // Don't try to write to a broken output
+    if (!os.good()  ||  m_OutputBroken) {
+        return -1;
+    }
+
     try {
         // HTTP header
         os << "Status: " << status_str << HTTP_EOL;
@@ -624,7 +732,6 @@ int CCgiApplication::OnException(exception& e, CNcbiOstream& os)
             }
         }
 
-
         // Check for problems in sending the response
         if ( !os.good() ) {
             ERR_POST_X(4, "CCgiApplication::OnException() failed to send error page"
@@ -666,6 +773,28 @@ const CArgs& CCgiApplication::GetArgs(void) const
 }
 
 
+class CExtraEntryCollector : public CEntryCollector_Base {
+public:
+    CExtraEntryCollector(void) {}
+    virtual ~CExtraEntryCollector(void) {}
+    
+    virtual void AddEntry(const string& name,
+                          const string& value,
+                          const string& filename,
+                          bool          is_index)
+    {
+        _ASSERT(!is_index  ||  value.empty());
+        m_Args.push_back(CDiagContext_Extra::TExtraArg(name,
+            filename.empty() ? value : filename + "/" + value));
+    }
+
+    CDiagContext_Extra::TExtraArgs& GetArgs(void) { return m_Args; }
+
+private:
+    CDiagContext_Extra::TExtraArgs m_Args;
+};
+
+
 void CCgiApplication::x_OnEvent(EEvent event, int status)
 {
     switch ( event ) {
@@ -676,29 +805,35 @@ void CCgiApplication::x_OnEvent(EEvent event, int status)
 
             // Print request start message
             if ( !CDiagContext::IsSetOldPostFormat() ) {
-                GetDiagContext().PrintRequestStart(req.GetCGIEntriesStr());
+                CExtraEntryCollector collector;
+                req.GetCGIEntries(collector);
+                GetDiagContext().PrintRequestStart()
+                    .AllowBadSymbolsInArgNames()
+                    .Print(collector.GetArgs());
                 m_RequestStartPrinted = true;
             }
 
             // Set default HTTP status code (reset above by PrintRequestStart())
             SetHTTPStatus(200);
+            m_ErrorStatus = false;
+
+            // This will log ncbi_phid as a separate 'extra' message
+            // if not yet logged.
+            CDiagContext::GetRequestContext().GetHitID();
 
-            const string& phid = CDiagContext::GetRequestContext().GetHitID();
             // Check if ncbi_st cookie is set
             const CCgiCookie* st = req.GetCookies().Find(
                 g_GetNcbiString(eNcbiStrings_Stat));
-            CUrlArgs pg_info;
             if ( st ) {
-                pg_info.SetQueryString(st->GetValue());
+                CUrlArgs pg_info(st->GetValue());
+                // Log ncbi_st values
+                CDiagContext_Extra extra = GetDiagContext().Extra();
+                // extra.SetType("NCBICGI");
+                ITERATE(CUrlArgs::TArgs, it, pg_info.GetArgs()) {
+                    extra.Print(it->name, it->value);
+                }
+                extra.Flush();
             }
-            pg_info.SetValue(g_GetNcbiString(eNcbiStrings_PHID), phid);
-            // Log ncbi_st values
-            CDiagContext_Extra extra = GetDiagContext().Extra();
-            // extra.SetType("NCBICGI");
-            ITERATE(CUrlArgs::TArgs, it, pg_info.GetArgs()) {
-                extra.Print(it->name, it->value);
-            }
-            extra.Flush();
             break;
         }
     case eSuccess:
@@ -706,23 +841,60 @@ void CCgiApplication::x_OnEvent(EEvent event, int status)
     case eException:
         {
             CRequestContext& rctx = GetDiagContext().GetRequestContext();
-            if ( m_InputStream.get() ) {
-                if ( m_InputStream->eof() ) {
-                    m_InputStream->clear();
+            try {
+                if ( m_InputStream.get() ) {
+                    if ( !m_InputStream->good() ) {
+                        m_InputStream->clear();
+                    }
+                    rctx.SetBytesRd(NcbiStreamposToInt8(m_InputStream->tellg()));
                 }
-                rctx.SetBytesRd(NcbiStreamposToInt8(m_InputStream->tellg()));
             }
-            if ( m_OutputStream.get() ) {
-                rctx.SetBytesWr(NcbiStreamposToInt8(m_OutputStream->tellp()));
+            catch (exception&) {
+            }
+            try {
+                if ( m_OutputStream.get() ) {
+                    if ( !m_OutputStream->good() ) {
+                        m_OutputBroken = true; // set flag to indicate broken output
+                        m_OutputStream->clear();
+                    }
+                    rctx.SetBytesWr(NcbiStreamposToInt8(m_OutputStream->tellp()));
+                }
+            }
+            catch (exception&) {
             }
             break;
         }
     case eEndRequest:
         {
+            CDiagContext& ctx = GetDiagContext();
+            CRequestContext& rctx = ctx.GetRequestContext();
+            // If an error status has been set by ProcessRequest, don't try
+            // to check the output stream and change the status to 299/499.
+            if ( !m_ErrorStatus ) {
+                // Log broken connection as 299/499 status
+                CNcbiOstream* os = m_Context.get() ?
+                    m_Context->GetResponse().GetOutput() : NULL;
+                if ((os  &&  !os->good())  ||  m_OutputBroken) {
+                    // 'Accept-Ranges: bytes' indicates a request for
+                    // content length, broken connection is OK.
+                    // If Content-Range is also set, the client was downloading
+                    // partial data. Broken connection is not OK in this case.
+                    if (TClientConnIntOk::GetDefault()  ||
+                        (m_Context->GetResponse().AcceptRangesBytes()  &&
+                        !m_Context->GetResponse().HaveContentRange())) {
+                        rctx.SetRequestStatus(
+                            CRequestStatus::e299_PartialContentBrokenConnection);
+                    }
+                    else {
+                        rctx.SetRequestStatus(
+                            CRequestStatus::e499_BrokenConnection);
+                    }
+                }
+            }
             if ( m_RequestStartPrinted  &&
                 !CDiagContext::IsSetOldPostFormat() ) {
                 // This will also reset request context
-                GetDiagContext().PrintRequestStop();
+                ctx.PrintRequestStop();
                 m_RequestStartPrinted = false;
             }
             break;
@@ -822,7 +994,7 @@ void CCgiApplication::ConfigureDiagFormat(CCgiContext& context)
     const CCgiRequest& request = context.GetRequest();
 
     typedef map<string, TDiagPostFlags> TFlagMap;
-    static CSafeStaticPtr<TFlagMap> s_FlagMap;
+    static CSafeStatic<TFlagMap> s_FlagMap;
     TFlagMap& flagmap = s_FlagMap.Get();
 
     TDiagPostFlags defaults = (eDPF_Prefix | eDPF_Severity
@@ -906,16 +1078,20 @@ CCgiStatistics* CCgiApplication::CreateStat()
     return new CCgiStatistics(*this);
 }
 
+
 ICgiSessionStorage* 
 CCgiApplication::GetSessionStorage(CCgiSessionParameters&) const 
 {
     return 0;
 }
+
+
 bool CCgiApplication::IsCachingNeeded(const CCgiRequest& request) const
 {
     return true;
 }
 
+
 ICache* CCgiApplication::GetCacheStorage() const
 { 
     return NULL;
@@ -952,6 +1128,8 @@ void CCgiApplication::SetRequestId(const string& rid, bool is_done)
     m_RID = rid;
     m_IsResultReady = is_done;
 }
+
+
 bool CCgiApplication::GetResultFromCache(const CCgiRequest& request, CNcbiOstream& os)
 {
     string checksum, content;
@@ -971,6 +1149,8 @@ bool CCgiApplication::GetResultFromCache(const CCgiRequest& request, CNcbiOstrea
     }
     return false;
 }
+
+
 void CCgiApplication::SaveResultToCache(const CCgiRequest& request, CNcbiIstream& is)
 {
     string checksum, content;
@@ -989,6 +1169,7 @@ void CCgiApplication::SaveResultToCache(const CCgiRequest& request, CNcbiIstream
     } 
 }
 
+
 void CCgiApplication::SaveRequest(const string& rid, const CCgiRequest& request)
 {    
     if (rid.empty())
@@ -1003,6 +1184,8 @@ void CCgiApplication::SaveRequest(const string& rid, const CCgiRequest& request)
         ERR_POST_X(7, "Couldn't save request : " << ex.what());
     } 
 }
+
+
 CCgiRequest* CCgiApplication::GetSavedRequest(const string& rid)
 {
     if (rid.empty())
@@ -1021,6 +1204,7 @@ CCgiRequest* CCgiApplication::GetSavedRequest(const string& rid)
     return NULL;
 }
 
+
 void CCgiApplication::x_AddLBCookie()
 {
     const CNcbiRegistry& reg = GetConfig();
@@ -1157,9 +1341,27 @@ string CCgiApplication::GetDefaultLogPath(void) const
 }
 
 
-void CCgiApplication::SetHTTPStatus(int status)
+void CCgiApplication::SetHTTPStatus(unsigned int status, const string& reason)
 {
-    CDiagContext::GetRequestContext().SetRequestStatus(status);
+    if ( m_Context.get() ) {
+        m_Context->GetResponse().SetStatus(status, reason);
+    }
+    else {
+        CDiagContext::GetRequestContext().SetRequestStatus(status);
+    }
+}
+
+
+bool CCgiApplication::x_DoneHeadRequest(void) const
+{
+    const CCgiContext& ctx = GetContext();
+    const CCgiRequest& req = ctx.GetRequest();
+    const CCgiResponse& res = ctx.GetResponse();
+    if (req.GetRequestMethod() != CCgiRequest::eMethod_HEAD  ||
+        !res.IsHeaderWritten() ) {
+        return false;
+    }
+    return true;
 }
 
 
@@ -1335,6 +1537,7 @@ string CCgiStatistics::Compose_ErrMessage(void)
 
 NCBI_PARAM_DEF(bool, CGI, DisableTrackingCookie, false);
 NCBI_PARAM_DEF(string, CGI, TrackingCookieName, "ncbi_sid");
+NCBI_PARAM_DEF(string, CGI, TrackingTagName, "NCBI-SID");
 NCBI_PARAM_DEF(string, CGI, TrackingCookieDomain, ".nih.gov");
 NCBI_PARAM_DEF(string, CGI, TrackingCookiePath, "/");
 
