@@ -1,4 +1,4 @@
-/*  $Id: reference_item.cpp 239986 2011-02-02 16:08:05Z rafanovi $
+/*  $Id: reference_item.cpp 425391 2014-01-28 20:03:15Z gotvyans $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
@@ -42,6 +42,7 @@
 #include <objects/general/Date.hpp>
 #include <objects/general/Date_std.hpp>
 #include <objects/medline/Medline_entry.hpp>
+#include <objects/mla/mla_client.hpp>
 #include <objects/pub/Pub.hpp>
 #include <objects/pub/Pub_equiv.hpp>
 #include <objects/pub/Pub_set.hpp>
@@ -54,6 +55,7 @@
 #include <objects/biblio/Imprint.hpp>
 #include <objects/submit/Submit_block.hpp>
 #include <objmgr/bioseq_ci.hpp>
+#include <objmgr/object_manager.hpp>
 #include <objmgr/util/sequence.hpp>
 #include <objmgr/util/seq_loc_util.hpp>
 
@@ -79,11 +81,15 @@ USING_SCOPE(sequence);
 class LessThan
 {
 public:
-    LessThan(bool serial_first, bool is_refseq);
+    enum ESerialFirst {
+        eSerialFirst_No = 0,
+        eSerialFirst_Yes
+    };
+    LessThan(ESerialFirst serial_first, bool is_refseq);
     bool operator()(const CRef<CReferenceItem>& ref1, const CRef<CReferenceItem>& ref2);
 private:
-    bool m_SerialFirst;
-    bool m_IsRefSeq;
+    ESerialFirst m_SerialFirst;
+    bool         m_IsRefSeq;
 };
 
 /////////////////////////////////////////////////////////////////////////////
@@ -325,7 +331,8 @@ static bool s_ShouldRemoveRef
 
 static void s_CombineRefs
 ( CReferenceItem& prev_ref,
-  CReferenceItem& curr_ref)
+  CReferenceItem& curr_ref,
+  CBioseqContext &ctx)
 {
     // merge locations
     {{
@@ -335,8 +342,12 @@ static void s_CombineRefs
         CConstRef<CSeq_id> prev_id( prev_loc->GetId() );
         CConstRef<CSeq_id> curr_id( curr_loc->GetId() );
 
+        const bool is_circular = 
+            ( ctx.GetHandle().CanGetInst_Topology() && ctx.GetHandle().GetInst_Topology() == CSeq_inst_Base::eTopology_circular );
+
         CRef<CSeq_loc> new_loc = Seq_loc_Add( *prev_loc, *curr_loc,
-            CSeq_loc::fMerge_All, &prev_ref.GetContext()->GetScope() );
+            ( is_circular ? CSeq_loc::fMerge_All : CSeq_loc::fSortAndMerge_All ),
+            &prev_ref.GetContext()->GetScope() );
 
         // save the old id because sometimes the merging changes it
         // We check for sameness to make sure it's reasonable to do this.
@@ -388,57 +399,10 @@ static void s_MergeDuplicates
         return;
     }
 
-    // see if merging is allowed (set sourcePubFuse)
-    bool sourcePubFuse = false;
-    {{
-        if( ctx.GetHandle().CanGetId() ) {
-            CBioseq_CI bioseq_iter( ctx.GetScope(), *ctx.GetTopLevelEntry().GetSeq_entryCore() );
-            for( ; bioseq_iter; ++bioseq_iter ) {
-                ITERATE( CBioseq_Handle::TId, it, bioseq_iter->GetId() )  { // old: ctx.GetHandle().GetId() , then .GetBioseqIds()
-                    CConstRef<CSeq_id> seqId = (*it).GetSeqIdOrNull();
-                    if( ! seqId.IsNull() ) {
-                        switch( seqId->Which() ) {
-                        case CSeq_id_Base::e_Gibbsq:
-                        case CSeq_id_Base::e_Gibbmt:
-                        case CSeq_id_Base::e_Embl:
-                        case CSeq_id_Base::e_Pir:
-                        case CSeq_id_Base::e_Swissprot:
-                        case CSeq_id_Base::e_Patent:        
-                        case CSeq_id_Base::e_Ddbj:
-                        case CSeq_id_Base::e_Prf:
-                        case CSeq_id_Base::e_Pdb:
-                        case CSeq_id_Base::e_Tpe:
-                        case CSeq_id_Base::e_Tpd:
-                        case CSeq_id_Base::e_Gpipe:
-                            // with some types, it's okay to merge
-                            sourcePubFuse = true;
-                            break;                        
-                        case CSeq_id_Base::e_Genbank:
-                        case CSeq_id_Base::e_Tpg:
-                            // Genbank allows merging only if it's the old-style 1 + 5 accessions
-                            if( NULL != seqId->GetTextseq_Id() &&
-                                seqId->GetTextseq_Id()->GetAccession().length() == 6 ) {
-                                    sourcePubFuse = true;
-                            }
-                            break;
-                        case CSeq_id_Base::e_not_set:
-                        case CSeq_id_Base::e_Local:
-                        case CSeq_id_Base::e_Other:
-                        case CSeq_id_Base::e_General:
-                        case CSeq_id_Base::e_Giim:                        
-                        case CSeq_id_Base::e_Gi:
-                            break;
-                        default:
-                            break;
-                        }
-                    }
-                }
-            }
-        }
-    }}       
-
-    // no merging allowed in this case
-    if( ! sourcePubFuse ) {
+    // see if merging is allowed
+    if( ! ctx.CanGetTLSeqEntryCtx() || 
+        ! ctx.GetTLSeqEntryCtx().GetCanSourcePubsBeFused() )
+    {
       return;
     }
 
@@ -488,7 +452,7 @@ static void s_MergeDuplicates
 
             if( s_ShouldRemoveRef( prev_ref, curr_ref ) ) {
                 if( combine_allowed ) {
-                    s_CombineRefs( prev_ref, curr_ref );
+                    s_CombineRefs( prev_ref, curr_ref, ctx );
                 }
                 remove = true;
             }
@@ -506,7 +470,8 @@ static void s_MergeDuplicates
 void CReferenceItem::Rearrange(TReferences& refs, CBioseqContext& ctx)
 {
     {{
-        stable_sort(refs.begin(), refs.end(), LessThan(false, ctx.IsRefSeq()));
+        stable_sort(refs.begin(), refs.end(), 
+            LessThan(LessThan::eSerialFirst_No, ctx.IsRefSeq()));
     }}
 
     {{
@@ -522,7 +487,8 @@ void CReferenceItem::Rearrange(TReferences& refs, CBioseqContext& ctx)
 
     {{
         // re-sort, take serial number into consideration.
-        stable_sort(refs.begin(), refs.end(), LessThan(true, ctx.IsRefSeq()));
+        stable_sort(refs.begin(), refs.end(), 
+            LessThan(LessThan::eSerialFirst_Yes, ctx.IsRefSeq()));
     }}
     
     // assign final serial numbers
@@ -570,8 +536,8 @@ static bool s_IsOnlySerial(const CPub& pub)
 
     const CCit_gen& gen = pub.GetGen();
 
-    if (!gen.IsSetCit()  ||
-        !NStr::StartsWith(gen.GetCit(), "BackBone id_pub", NStr::eNocase)) {
+    if ( !gen.IsSetCit() ) 
+    {
         if (!gen.IsSetJournal()  &&  !gen.IsSetDate()  &&
             gen.IsSetSerial_number()  &&  gen.GetSerial_number() > 0) {
             return true;
@@ -597,7 +563,7 @@ void CReferenceItem::x_CreateUniqueStr(void) const
             continue;
         }
         if (!s_IsOnlySerial(pub)) {
-            pub.GetLabel(&m_UniqueStr, CPub::eContent, true);
+            pub.GetLabel(&m_UniqueStr, CPub::eContent, CPub::fLabel_Unique, CPub::eLabel_V1 );
         }
     }
 }
@@ -627,7 +593,7 @@ bool CReferenceItem::Matches(const CPub& pub) const
                 const string& uniquestr = m_UniqueStr;
 
                 string pub_unique;
-                pub.GetLabel(&pub_unique, CPub::eContent, true);
+                pub.GetLabel(&pub_unique, CPub::eContent, CPub::fLabel_Unique, CPub::eLabel_V1 );
 
                 size_t len = pub_unique.length();
                 if (len > 0  &&  pub_unique[len - 1] == '>') {
@@ -671,6 +637,79 @@ void CReferenceItem::x_GatherInfo(CBioseqContext& ctx)
         x_Init(**it, ctx);
     }
 
+    // if just pmid or just muid, we look it up, assuming the user has
+    // somehow given permission for remote lookups 
+    const static string kGbLoader = "GBLOADER";
+    if( IsJustUids() && 
+        ctx.GetScope().GetObjectManager().FindDataLoader(kGbLoader) ) 
+    {
+        // TODO: To avoid repeated connections, in the future we should have 
+        // one CMLAClient shared by the whole program
+        CMLAClient mlaClient;
+
+        vector< CRef<CPub> > new_pubs;
+
+        ITERATE( CPub_equiv::Tdata, it, pub.Get() ) {
+            const CPub & pub = **it;
+            CRef<CPub> new_pub;
+
+            try {
+                switch(pub.Which()) {
+                case CPub::e_Pmid:
+                    {
+                        const int pmid = pub.GetPmid().Get();
+
+                        CPubMedId req(pmid);
+                        CMLAClient::TReply reply;
+                        new_pub = mlaClient.AskGetpubpmid(req, &reply);
+                    }
+                    break;
+                case CPub::e_Muid:
+                    {
+                        const int muid = pub.GetMuid();
+
+                        const int pmid = mlaClient.AskUidtopmid(muid);
+                        if( pmid > 0 ) {
+                            CPubMedId req(pmid);
+                            CMLAClient::TReply reply;
+                            new_pub = mlaClient.AskGetpubpmid(req, &reply);
+                        }
+                    }
+                    break;
+                default:
+                    // ignore if type unknown
+                    break;
+                }
+            } catch(...) {
+                // don't worry if we can't look it up
+            }
+
+            if( new_pub ) {
+                // authors come back in a weird format that we need
+                // to convert to ISO
+                ChangeMedlineAuthorsToISO(new_pub);
+
+                new_pubs.push_back(new_pub);
+            }
+        }
+
+        if( ! new_pubs.empty() ) {
+            ITERATE( vector< CRef<CPub> >, new_pub_iter, new_pubs ) {
+                x_Init( **new_pub_iter, ctx );
+            }
+
+            // we have to add the new_pubs to m_Pubdesc->GetPub() but m_Pubdesc
+            // is const.  The solution is to copy it, modify the copy, and 
+            // set the copy to have CConstRef
+            CRef<CPubdesc> new_pubdesc( new CPubdesc );
+            new_pubdesc->Assign(*m_Pubdesc);
+            CPub_equiv::Tdata & new_pub_list = new_pubdesc->SetPub().Set();
+            copy( new_pubs.begin(), new_pubs.end(),
+                back_inserter(new_pub_list) );
+            m_Pubdesc = new_pubdesc;
+        }
+    }
+
     // gather Genbank specific fields (formats: Genbank, GBSeq, DDBJ)
     if ( ctx.IsGenbankFormat() ) {
         x_GatherRemark(ctx);
@@ -968,6 +1007,9 @@ void CReferenceItem::x_InitProc(const CCit_book& proc, CBioseqContext& ctx)
     if (proc.IsSetTitle()) {
         m_Title = proc.GetTitle().GetTitle();
     }
+    if (proc.CanGetImp()) {
+        x_AddImprint(proc.GetImp(), ctx);
+    }
 }
 
 
@@ -982,6 +1024,7 @@ void CReferenceItem::x_Init
     if (book.CanGetImp()) {
         x_AddImprint(book.GetImp(), ctx);
     }
+
 }
 
 
@@ -1135,11 +1178,11 @@ void CReferenceItem::FormatAuthors(const CAuth_list& alp, string& auth)
     ITERATE (TStrList, it, authors) {
         auth_line << separator << *it;
         ++it;
-        // TODO: might want to remove "et al" detection once we've moved over to the C++ toolkit.
+        // might want to remove "et al" detection once we've moved over to the C++ toolkit.
         // It's here to make the diffs match.
-        if( (it != authors.end()) && 
+        if( ( it == last ) &&
             ( NStr::StartsWith(*it, "et al", NStr::eNocase) || NStr::StartsWith(*it, "et,al", NStr::eNocase) ) ) {
-                separator = ", ";
+                separator = " ";
         } else if (it == last) {
             separator = " and ";
         } else {
@@ -1182,10 +1225,25 @@ void CReferenceItem::x_CleanData(void)
     s_RemovePeriod(m_Title);
     x_CapitalizeTitleIfNecessary();
     // remark
-    ConvertQuotes(m_Remark);
+    ConvertQuotesNotInHTMLTags(m_Remark);
     ExpandTildes(m_Remark, eTilde_newline);
 }
 
+// Unfortunately, some compilers won't let us pass
+// islower straight to STL algorithms
+class CIsLowercase {
+public:
+    bool operator()( const char ch ) {
+        return 0 != islower(ch);
+    }
+};
+
+static
+bool s_ContainsNoLowercase( const string &str ) 
+{
+    return ( find_if( str.begin(), str.end(), CIsLowercase() ) == str.end() );
+}
+
 void CReferenceItem::x_CapitalizeTitleIfNecessary()
 {
     if( ! GetPubdesc().CanGetPub() ) {
@@ -1202,8 +1260,15 @@ void CReferenceItem::x_CapitalizeTitleIfNecessary()
         switch( pub.Which() ) {
             case CPub::e_Proc:
             case CPub::e_Man:
-                if(  ! m_Title.empty() ) {
-                    m_Title[0] = toupper( m_Title[0] ); // capitalize the title
+                if(  m_Title.length() > 3 ) {
+                    // capitalize the title before checking for all caps
+                    m_Title[0] = toupper( m_Title[0] );
+                    if( s_ContainsNoLowercase(m_Title) ) {
+                        NStr::ToLower( m_Title );
+                        // we undid the earlier uppercasing step.  Slightly slower, but
+                        // the code is a little cleaner than using a loop
+                        m_Title[0] = toupper( m_Title[0] );
+                    }
                     return;
                 }
                 break;
@@ -1214,6 +1279,121 @@ void CReferenceItem::x_CapitalizeTitleIfNecessary()
     }
 }
 
+void CReferenceItem::ChangeMedlineAuthorsToISO( CRef<CPub> pub )
+{
+    // leave early if it doesn't need to be changed
+    if( ! pub || ! pub->IsArticle() || ! pub->IsSetAuthors() || ! pub->GetAuthors().IsSetNames() ||
+        ! pub->GetAuthors().GetNames().IsMl() )
+    {
+        return;
+    }
+
+    // build our new authors list in here
+    CAuth_list::C_Names::TStd new_authors;
+
+    const CAuth_list::C_Names::TMl & ml_names = pub->GetAuthors().GetNames().GetMl();
+    ITERATE( CAuth_list::C_Names::TMl, ml_name_iter, ml_names ) {
+        string author_name = *ml_name_iter;
+
+        // we will fill in these 3 as we go along
+        string lastname;
+        string initials;
+        string suffix;
+
+        // this scope fills in lastname, initials, and suffix
+        {
+            NStr::TruncateSpacesInPlace(author_name);
+            vector<string> tokens;
+            NStr::Tokenize(author_name, " ", tokens, NStr::eMergeDelims);
+
+            // get suffix if it exists, and remove it from the list
+            if( tokens.size() >= 3 && 
+                ! x_StringIsJustCapitalLetters(tokens.back()) && 
+                x_StringIsJustCapitalLetters(tokens[tokens.size() - 2]) ) 
+            {
+                suffix = tokens.back();
+                tokens.pop_back();
+            }
+
+            // get initials if they exist, and remove them from the list
+            if( tokens.size() >= 2 &&
+                x_StringIsJustCapitalLetters(tokens.back()) )
+            {
+                initials = tokens.back();
+                tokens.pop_back();
+            }
+
+            // remaining pieces belong to the last name
+            lastname = NStr::Join(tokens, " ");
+        }
+
+        // put period in initials. e.g. "MJ" -> "M.J."
+        {
+            string new_initials;
+            ITERATE( string, ch_iter, initials ) {
+                new_initials += *ch_iter;
+                new_initials += '.';
+            }
+            // swap is faster than assignment
+            initials.swap(new_initials);
+        }
+
+        // a couple of static transformations for the suffix
+        typedef SStaticPair<const char*, const char*> TSufElem;
+        static const TSufElem sc_suf_map[] = {
+            { "1d",  "I" },
+            { "1st", "I" },
+            { "2d",  "II" },
+            { "2nd", "II" },
+            { "3d",  "III" },
+            { "3rd", "III" },
+            { "4th", "IV" },
+            { "5th", "V" },
+            { "6th", "VI" },
+            { "Jr",  "Jr." },
+            { "Sr",  "Sr."}
+        };
+        typedef CStaticArrayMap<const char *, const char *, PCase_CStr> TSufMap;
+        DEFINE_STATIC_ARRAY_MAP(TSufMap, sc_SufMap, sc_suf_map);
+
+        TSufMap::const_iterator suf_find_iter = sc_SufMap.find(suffix.c_str());
+        if( suf_find_iter != sc_SufMap.end() ) {
+            suffix = suf_find_iter->second;
+        }
+
+        CRef<CAuthor> new_author( new CAuthor );
+        CPerson_id_Base::TName & name = new_author->SetName().SetName();
+        name.SetLast( lastname );
+        if( ! initials.empty() ) {
+            name.SetInitials( initials );
+        }
+        if( ! suffix.empty() ) {
+            name.SetSuffix( suffix );
+        }
+
+        new_authors.push_back( new_author );
+    }
+    
+    copy( new_authors.begin(), 
+        new_authors.end(),
+        back_inserter( pub->SetArticle().SetAuthors().SetNames().SetStd() ) );
+}
+
+bool CReferenceItem::x_StringIsJustCapitalLetters( const string & str )
+{
+    if( str.empty() ) {
+        return false;
+    }
+
+    ITERATE(string, ch_iter, str ) {
+        if( ! isupper(*ch_iter) ) {
+            return false;
+        }
+    }
+
+    return true;
+}
+
 
 /////////////////////////////////////////////////////////////////////////////
 //
@@ -1221,7 +1401,7 @@ void CReferenceItem::x_CapitalizeTitleIfNecessary()
 
 // these must be in "ASCIIbetical" order; beware of the fact that
 // closing quotes sort after spaces.
-static const string sc_RemarkText[] = {
+static const char* const sc_RemarkText[] = {
   "full automatic",
   "full staff_entry",
   "full staff_review",
@@ -1232,11 +1412,14 @@ static const string sc_RemarkText[] = {
   "unannotated staff_entry",
   "unannotated staff_review"
 };
-DEFINE_STATIC_ARRAY_MAP(CStaticArraySet<string>, sc_Remarks, sc_RemarkText);
+typedef CStaticArraySet<const char*, PCase_CStr> TStaticRemarkSet;
+DEFINE_STATIC_ARRAY_MAP(TStaticRemarkSet, sc_Remarks, sc_RemarkText);
 
 
 void CReferenceItem::x_GatherRemark(CBioseqContext& ctx)
 {
+    static const char* const kDoiLink = "http://dx.doi.org/";
+
     list<string> l;
 
     // comment
@@ -1244,7 +1427,7 @@ void CReferenceItem::x_GatherRemark(CBioseqContext& ctx)
     if ( m_Pubdesc->IsSetComment()  &&  !m_Pubdesc->GetComment().empty() ) {
         const string& comment = m_Pubdesc->GetComment();
         
-        if ( sc_Remarks.find(comment) == sc_Remarks.end() ) {
+        if ( sc_Remarks.find(comment.c_str()) == sc_Remarks.end() ) {
             l.push_back(comment);
         }
     }
@@ -1276,6 +1459,7 @@ void CReferenceItem::x_GatherRemark(CBioseqContext& ctx)
             if (!NStr::EndsWith(l.back(), '.')) {
                 AddPeriod(l.back());
             }
+            NStr::ReplaceInPlace( l.back(), "\"", "\'" );
         }
 
         // Poly_a
@@ -1307,6 +1491,8 @@ void CReferenceItem::x_GatherRemark(CBioseqContext& ctx)
                                 if (ret.IsSetExp()  &&
                                     !ret.GetExp().empty() ) {
                                     l.push_back("Erratum:[" + ret.GetExp() + "]");
+                                } else {
+                                    l.push_back("Erratum");
                                 }
                                 break;
 
@@ -1314,6 +1500,8 @@ void CReferenceItem::x_GatherRemark(CBioseqContext& ctx)
                                 if (ret.IsSetExp()  &&
                                     !ret.GetExp().empty() ) {
                                     l.push_back("Retracted:[" + ret.GetExp() + "]");
+                                } else {
+                                    l.push_back("Retracted");
                                 }
                                 break;
 
@@ -1321,6 +1509,8 @@ void CReferenceItem::x_GatherRemark(CBioseqContext& ctx)
                                 if (ret.IsSetExp()  &&
                                     !ret.GetExp().empty() ) {
                                     l.push_back("Correction to:[" + ret.GetExp() + "]");
+                                } else {
+                                    l.push_back("Correction");
                                 }
                                 break;
 
@@ -1365,7 +1555,15 @@ void CReferenceItem::x_GatherRemark(CBioseqContext& ctx)
                                 if( (*it)->Which() == CArticleId_Base::e_Doi) {
                                     const string & doi = (*it)->GetDoi().Get();
                                     if( NStr::StartsWith( doi, "10." ) ) {
-                                        l.push_back( "DOI: " + doi );
+                                        if( ctx.Config().DoHTML() && ! CommentHasSuspiciousHtml(doi) ) {
+                                            CNcbiOstrstream result;
+                                            result << "DOI: <a href=\""
+                                                   << kDoiLink << doi << "\">"
+                                                   << doi << "</a>";
+                                            l.push_back( CNcbiOstrstreamToString(result) );
+                                        } else {
+                                            l.push_back( "DOI: " + doi );
+                                        }
                                         break;
                                     }
                                 }
@@ -1458,7 +1656,7 @@ static CDate::ECompare s_CompareDates(const CDate& d1, const CDate& d2)
     return CDate::eCompare_same;
 }
 
-LessThan::LessThan(bool serial_first, bool is_refseq) :
+LessThan::LessThan(ESerialFirst serial_first, bool is_refseq) :
     m_SerialFirst(serial_first), m_IsRefSeq(is_refseq)
 {}
 
@@ -1467,7 +1665,7 @@ bool LessThan::operator()
 (const CRef<CReferenceItem>& ref1,
  const CRef<CReferenceItem>& ref2)
 {
-    if ( m_SerialFirst  &&  ref1->GetSerial() != ref2->GetSerial() ) {
+    if ( m_SerialFirst == eSerialFirst_Yes &&  ref1->GetSerial() != ref2->GetSerial() ) {
         return ref1->GetSerial() < ref2->GetSerial();
     }
 
@@ -1528,9 +1726,21 @@ bool LessThan::operator()
     string auth1, auth2;
     if (ref1->IsSetAuthors()) {
         CReferenceItem::FormatAuthors(ref1->GetAuthors(), auth1);
+        if( ! ref1->GetConsortium().empty() ) {
+            if( ! auth1.empty() ) {
+                auth1 += "; ";
+            }
+            auth1 += ref1->GetConsortium();
+        }
     }
     if (ref2->IsSetAuthors()) {
         CReferenceItem::FormatAuthors(ref2->GetAuthors(), auth2);
+        if( ! ref2->GetConsortium().empty() ) {
+            if( ! auth2.empty() ) {
+                auth2 += "; ";
+            }
+            auth2 += ref2->GetConsortium();
+        }
     }
     int comp = NStr::CompareNocase(auth1, auth2);
     if ( comp != 0 ) {
