@@ -1,4 +1,4 @@
-/*  $Id: xsdparser.cpp 203020 2010-08-31 13:13:48Z gouriano $
+/*  $Id: xsdparser.cpp 432804 2014-04-18 13:01:30Z gouriano $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
@@ -132,6 +132,7 @@ void XSDParser::Reset(void)
     m_NamespaceToPrefix.clear();
     m_TargetNamespace.erase();
     m_ElementFormDefault = false;
+    m_AttributeFormDefault = false;
 }
 
 TToken XSDParser::GetNextToken(void)
@@ -248,9 +249,13 @@ bool XSDParser::IsValue(const char* value) const
 
 bool XSDParser::DefineElementType(DTDElement& node)
 {
+    if (GetElementNamespace(m_ValuePrefix) != eSchemaNamespace) {
+        return false;
+    }
     if (IsValue("string") || IsValue("token") ||
         IsValue("normalizedString") ||
-        IsValue("anyURI") || IsValue("QName") ||
+        IsValue("duration") || IsValue("language") ||
+        IsValue("anyType") || IsValue("anyURI") || IsValue("QName") ||
         IsValue("dateTime") || IsValue("time") || IsValue("date") ||
         IsValue("anySimpleType")) {
         node.SetType(DTDElement::eString);
@@ -279,8 +284,12 @@ bool XSDParser::DefineElementType(DTDElement& node)
 
 bool XSDParser::DefineAttributeType(DTDAttribute& attrib)
 {
+    if (GetElementNamespace(m_ValuePrefix) != eSchemaNamespace) {
+        return false;
+    }
     if (IsValue("string") || IsValue("token") || IsValue("QName") ||
-        IsValue("anyURI") ||
+        IsValue("language") ||
+        IsValue("anyType") || IsValue("anyURI") ||
         IsValue("dateTime") || IsValue("time") || IsValue("date")) {
         attrib.SetType(DTDAttribute::eString);
     } else if (IsValue("ID")) {
@@ -343,6 +352,8 @@ void XSDParser::ParseHeader()
                 m_TargetNamespace = m_Value;
             } else if (IsAttribute("elementFormDefault")) {
                 m_ElementFormDefault = IsValue("qualified");
+            } else if (IsAttribute("attributeFormDefault")) {
+                m_AttributeFormDefault = IsValue("qualified");
             }
         }
     }
@@ -367,19 +378,20 @@ void XSDParser::ParseInclude(void)
     if (name.empty()) {
         ParseError("schemaLocation");
     }
-    DTDEntity& node = m_MapEntity[CreateEntityId(name, DTDEntity::eEntity)];
+    string id(CreateEntityId(name, DTDEntity::eEntity));
+    DTDEntity& node = m_MapEntity[id];
     if (node.GetName().empty()) {
         node.SetName(name);
         node.SetData(name);
         node.SetExternal();
-        PushEntityLexer(name);
-        Reset();
+        PushEntityLexer(id);
         ParseHeader();
     }
 }
 
 void XSDParser::ParseImport(void)
 {
+    bool import=true;
     TToken tok = GetRawAttributeSet();
     if (GetAttribute("namespace")) {
         if (IsValue("http://www.w3.org/XML/1998/namespace") ||
@@ -387,8 +399,25 @@ void XSDParser::ParseImport(void)
             string name = "xml:lang";
             m_MapAttribute[name].SetName(name);
             m_MapAttribute[name].SetType(DTDAttribute::eString);
+            name = "xml:space";
+            m_MapAttribute[name].SetName(name);
+            m_MapAttribute[name].SetType(DTDAttribute::eString);
+            import=false;
         }
     }
+    if (import && GetAttribute("schemaLocation")) {
+        string name(m_Value);
+        string id(CreateEntityId(name, DTDEntity::eEntity));
+        DTDEntity& node = m_MapEntity[id];
+        if (node.GetName().empty()) {
+            node.SetName(name);
+            node.SetData(name);
+            node.SetExternal();
+            PushEntityLexer(id);
+            ParseHeader();
+        }
+        return;
+    }
     if (tok == K_CLOSING) {
         SkipContent();
     }
@@ -534,8 +563,9 @@ DTDElement::EOccurrence XSDParser::ParseMaxOccurs( DTDElement::EOccurrence occNo
 string XSDParser::ParseElementContent(DTDElement* owner, int emb)
 {
     TToken tok;
-    string name, value;
+    string name, value, name_space;
     bool ref=false, named_type=false;
+    bool qualified = m_ElementFormDefault;
     int line = Lexer().CurrentLine();
 
     tok = GetRawAttributeSet();
@@ -551,13 +581,16 @@ string XSDParser::ParseElementContent(DTDElement* owner, int emb)
             elem.SetType(DTDElement::eAny);
             ref=false;
         } else {
-            name = m_Value;
             ref=true;
+            name_space = m_ResolveTypes ? m_TargetNamespace :
+                                          m_PrefixToNamespace[m_ValuePrefix];
+            name = m_Value + name_space;
         }
     }
     if (GetAttribute("name")) {
         ref=false;
-        name = m_Value;
+        name_space = m_TargetNamespace;
+        name = m_Value + name_space;
         if (owner) {
             name = CreateTmpEmbeddedName(owner->GetName(), emb);
             m_MapElement[name].SetEmbedded();
@@ -569,13 +602,20 @@ string XSDParser::ParseElementContent(DTDElement* owner, int emb)
     }
     if (GetAttribute("type")) {
         if (!DefineElementType(m_MapElement[name])) {
-            m_MapElement[name].SetTypeName(m_Value);
+            m_MapElement[name].SetTypeName(
+                CreateEntityId(m_Value,DTDEntity::eType, &m_ValuePrefix));
             named_type = true;
         }
     }
+    if (owner && GetAttribute("form")) {
+        qualified = IsValue("qualified");
+    }
     if (GetAttribute("default")) {
         m_MapElement[name].SetDefault(m_Value);
     }
+    if (GetAttribute("nillable")) {
+        m_MapElement[name].SetNillable(IsValue("true"));
+    }
     if (owner && !name.empty()) {
         owner->SetOccurrence(name, ParseMinOccurs( owner->GetOccurrence(name)));
         owner->SetOccurrence(name, ParseMaxOccurs( owner->GetOccurrence(name)));
@@ -583,16 +623,20 @@ string XSDParser::ParseElementContent(DTDElement* owner, int emb)
     if (tok != K_CLOSING && tok != K_ENDOFTAG) {
         ParseError("endoftag");
     }
+    m_MapElement[name].SetNamespaceName(name_space);
+    m_MapElement[name].SetQualified(qualified);
     bool hasContents = false;
     if (tok == K_CLOSING) {
         hasContents = ParseContent(m_MapElement[name]);
     }
     m_ExpectLastComment = true;
     if (!ref && !named_type) {
-        m_MapElement[name].SetTypeIfUnknown(
-            hasContents ? DTDElement::eEmpty : DTDElement::eString);
+        named_type = NStr::StartsWith( m_MapElement[name].GetTypeName(), "type:");
+        if (!named_type) {
+            m_MapElement[name].SetTypeIfUnknown(
+                hasContents ? DTDElement::eEmpty : DTDElement::eString);
+        }
     }
-    m_MapElement[name].SetNamespaceName(m_TargetNamespace);
     return name;
 }
 
@@ -602,24 +646,21 @@ string XSDParser::ParseGroup(DTDElement* owner, int emb)
     TToken tok = GetRawAttributeSet();
     if (GetAttribute("ref")) {
 
-        string id = CreateEntityId(m_Value,DTDEntity::eGroup);
+        string id( CreateEntityId(m_Value,DTDEntity::eGroup,&m_ValuePrefix));
         name = CreateTmpEmbeddedName(owner->GetName(), emb);
         DTDElement& node = m_MapElement[name];
         node.SetEmbedded();
         node.SetName(m_Value);
         node.SetOccurrence( ParseMinOccurs( node.GetOccurrence()));
         node.SetOccurrence( ParseMaxOccurs( node.GetOccurrence()));
+        node.SetQualified(owner->IsQualified());
         SetCommentsIfEmpty(&(node.Comments()));
 
         if (m_ResolveTypes) {
-            if (m_MapEntity.find(id) != m_MapEntity.end()) {
-                PushEntityLexer(id);
-                ParseGroupRef(node);
-            } else {
-                ParseError("Unresolved entity", id.c_str());
-            }
+            PushEntityLexer(id);
+            ParseGroupRef(node);
         } else {
-            node.SetTypeName(node.GetName());
+            node.SetTypeName(id);
             node.SetType(DTDElement::eUnknownGroup);
             Lexer().FlushCommentsTo(node.Comments());
         }
@@ -694,6 +735,7 @@ bool XSDParser::ParseContent(DTDElement& node, bool extended /*=false*/)
                 elem.SetSourceLine(Lexer().CurrentLine());
                 elem.SetEmbedded();
                 elem.SetType(DTDElement::eAny);
+                elem.SetQualified(node.IsQualified());
                 ParseAny(elem);
                 AddElementContent(node,name);
             }
@@ -725,6 +767,7 @@ bool XSDParser::ParseContent(DTDElement& node, bool extended /*=false*/)
                 elem.SetSourceLine(Lexer().CurrentLine());
                 elem.SetEmbedded();
                 elem.SetType(DTDElement::eSequence);
+                elem.SetQualified(node.IsQualified());
                 ParseContainer(elem);
                 AddElementContent(node,name);
             }
@@ -746,6 +789,7 @@ bool XSDParser::ParseContent(DTDElement& node, bool extended /*=false*/)
                 elem.SetSourceLine(Lexer().CurrentLine());
                 elem.SetEmbedded();
                 elem.SetType(DTDElement::eChoice);
+                elem.SetQualified(node.IsQualified());
                 ParseContainer(elem);
                 AddElementContent(node,name);
             }
@@ -767,6 +811,7 @@ bool XSDParser::ParseContent(DTDElement& node, bool extended /*=false*/)
                 elem.SetSourceLine(Lexer().CurrentLine());
                 elem.SetEmbedded();
                 elem.SetType(DTDElement::eSet);
+                elem.SetQualified(node.IsQualified());
                 ParseContainer(elem);
                 AddElementContent(node,name);
             }
@@ -787,6 +832,12 @@ bool XSDParser::ParseContent(DTDElement& node, bool extended /*=false*/)
             SetCommentsIfEmpty(&(node.Comments()));
             ParseAnnotation();
             break;
+        case K_UNION:
+            ParseUnion(node);
+            break;
+        case K_LIST:
+            ParseList(node);
+            break;
         default:
             for ( tok = GetNextToken(); tok == K_ATTPAIR || tok == K_XMLNS; tok = GetNextToken())
                 ;
@@ -839,17 +890,13 @@ void XSDParser::ParseExtension(DTDElement& node)
     bool extended=false;
     if (GetAttribute("base")) {
         if (!DefineElementType(node)) {
+            string id( CreateEntityId(m_Value,DTDEntity::eType,&m_ValuePrefix));
             if (m_ResolveTypes) {
-                string id = CreateEntityId(m_Value,DTDEntity::eType);
-                if (m_MapEntity.find(id) != m_MapEntity.end()) {
-                    PushEntityLexer(id);
-                    ParseContent(node);
-                    extended=true;
-                } else {
-                    ParseError("Unresolved entity", id.c_str());
-                }
+                PushEntityLexer(id);
+                ParseContent(node);
+                extended=true;
             } else {
-                node.SetTypeName(m_Value);
+                node.SetTypeName(id);
             }
         }
     }
@@ -864,17 +911,13 @@ void XSDParser::ParseRestriction(DTDElement& node)
     bool extended=false;
     if (GetAttribute("base")) {
         if (!DefineElementType(node)) {
-            string id = CreateEntityId(m_Value,DTDEntity::eType);
+            string id( CreateEntityId(m_Value,DTDEntity::eType,&m_ValuePrefix));
             if (m_ResolveTypes) {
-                if (m_MapEntity.find(id) != m_MapEntity.end()) {
-                    PushEntityLexer(id);
-                    ParseContent(node);
-                    extended=true;
-                } else {
-                    ParseError("Unresolved entity", id.c_str());
-                }
+                PushEntityLexer(id);
+                ParseContent(node);
+                extended=true;
             } else {
-                node.SetTypeName(m_Value);
+                node.SetTypeName(id);
             }
         }
     }
@@ -891,10 +934,14 @@ void XSDParser::ParseAttribute(DTDElement& node)
     att.SetSourceLine(Lexer().CurrentLine());
     SetCommentsIfEmpty(&(att.Comments()));
     bool ref=false, named_type=false;
+    bool qualified = m_AttributeFormDefault;
 
     TToken tok = GetRawAttributeSet();
     if (GetAttribute("ref")) {
         att.SetName(m_Value);
+        if (m_ValuePrefix == "xml") {
+            att.SetName(m_ValuePrefix + ":" + m_Value);
+        }
         ref=true;
     }
     if (GetAttribute("name")) {
@@ -902,7 +949,8 @@ void XSDParser::ParseAttribute(DTDElement& node)
     }
     if (GetAttribute("type")) {
         if (!DefineAttributeType(att)) {
-            att.SetTypeName(m_Value);
+            att.SetTypeName(
+                CreateEntityId(m_Value,DTDEntity::eType,&m_ValuePrefix));
             named_type = true;
         }
     }
@@ -918,6 +966,10 @@ void XSDParser::ParseAttribute(DTDElement& node)
     if (GetAttribute("default")) {
         att.SetValue(m_Value);
     }
+    if (GetAttribute("form")) {
+        qualified = IsValue("qualified");
+    }
+    att.SetQualified(qualified);
     if (tok == K_CLOSING) {
         ParseContent(att);
     }
@@ -931,18 +983,14 @@ void XSDParser::ParseAttributeGroup(DTDElement& node)
 {
     TToken tok = GetRawAttributeSet();
     if (GetAttribute("ref")) {
+        string id( CreateEntityId(m_Value,DTDEntity::eAttGroup,&m_ValuePrefix));
         if (m_ResolveTypes) {
-            string id = CreateEntityId(m_Value,DTDEntity::eAttGroup);
-            if (m_MapEntity.find(id) != m_MapEntity.end()) {
-                PushEntityLexer(id);
-                ParseAttributeGroupRef(node);
-            } else {
-                ParseError("Unresolved entity", id.c_str());
-            }
+            PushEntityLexer(id);
+            ParseAttributeGroupRef(node);
         } else {
             DTDAttribute a;
             a.SetType(DTDAttribute::eUnknownGroup);
-            a.SetTypeName(m_Value);
+            a.SetTypeName(id);
             Lexer().FlushCommentsTo(node.AttribComments());
             node.AddAttribute(a);
         }
@@ -1027,9 +1075,11 @@ string XSDParser::ParseAttributeContent()
     }
     if (GetAttribute("type")) {
         if (!DefineAttributeType(m_MapAttribute[name])) {
-            m_MapAttribute[name].SetTypeName(m_Value);
+            m_MapAttribute[name].SetTypeName(
+                CreateEntityId(m_Value, DTDEntity::eType,&m_ValuePrefix));
         }
     }
+    m_MapAttribute[name].SetQualified(true);
     if (tok == K_CLOSING) {
         ParseContent(m_MapAttribute[name]);
     }
@@ -1078,16 +1128,12 @@ void XSDParser::ParseExtension(DTDAttribute& att)
     TToken tok = GetRawAttributeSet();
     if (GetAttribute("base")) {
         if (!DefineAttributeType(att)) {
+            string id( CreateEntityId(m_Value,DTDEntity::eType,&m_ValuePrefix));
             if (m_ResolveTypes) {
-                string id = CreateEntityId(m_Value,DTDEntity::eType);
-                if (m_MapEntity.find(id) != m_MapEntity.end()) {
-                    PushEntityLexer(id);
-                    ParseContent(att);
-                } else {
-                    ParseError("Unresolved entity", id.c_str());
-                }
+                PushEntityLexer(id);
+                ParseContent(att);
             } else {
-                att.SetTypeName(m_Value);
+                att.SetTypeName(id);
             }
         }
     }
@@ -1101,16 +1147,12 @@ void XSDParser::ParseRestriction(DTDAttribute& att)
     TToken tok = GetRawAttributeSet();
     if (GetAttribute("base")) {
         if (!DefineAttributeType(att)) {
+            string id( CreateEntityId(m_Value,DTDEntity::eType,&m_ValuePrefix));
             if (m_ResolveTypes) {
-                string id = CreateEntityId(m_Value,DTDEntity::eType);
-                if (m_MapEntity.find(id) != m_MapEntity.end()) {
-                    PushEntityLexer(id);
-                    ParseContent(att);
-                } else {
-                    ParseError("Unresolved entity", id.c_str());
-                }
+                PushEntityLexer(id);
+                ParseContent(att);
             } else {
-                att.SetTypeName(m_Value);
+                att.SetTypeName(id);
             }
         }
     }
@@ -1184,7 +1226,8 @@ string XSDParser::CreateTmpEmbeddedName(const string& name, int emb)
     return emb_name;
 }
 
-string XSDParser::CreateEntityId( const string& name, DTDEntity::EType type)
+string XSDParser::CreateEntityId(
+    const string& name, DTDEntity::EType type, const string* prefix)
 {
     string id;
     switch (type) {
@@ -1207,6 +1250,16 @@ string XSDParser::CreateEntityId( const string& name, DTDEntity::EType type)
         id = name;
         break;
     }
+    if (prefix) {
+        if (m_PrefixToNamespace.find(*prefix) == m_PrefixToNamespace.end()) {
+            string msg("Namespace prefix not defined: ");
+            msg += *prefix;
+            ParseError(msg.c_str(), "namespace declaration");
+        }
+        id += m_PrefixToNamespace[*prefix];
+    } else {
+        id += m_TargetNamespace;
+    }
     return id;
 }
 
@@ -1219,7 +1272,7 @@ void XSDParser::CreateTypeDefinition(DTDEntity::EType type)
         data += " " + m_Raw;
         if (IsAttribute("name")) {
             name = m_Value;
-            id = CreateEntityId(name,type);
+            id = CreateEntityId(m_Value,type);
             m_MapEntity[id].SetName(name);
         }
     }
@@ -1229,6 +1282,9 @@ void XSDParser::CreateTypeDefinition(DTDEntity::EType type)
     }
     m_MapEntity[id].SetData(data);
     m_MapEntity[id].SetType(type);
+    m_MapEntity[id].SetParseAttributes( m_TargetNamespace,
+        m_ElementFormDefault,m_AttributeFormDefault,
+        m_PrefixToNamespace);
     if (tok == K_CLOSING) {
         ParseTypeDefinition(m_MapEntity[id]);
     }
@@ -1336,32 +1392,84 @@ void XSDParser::ProcessNamedTypes(void)
                     if (j != m_MapElement.end()) {
                         break;
                     }
-                    PushEntityLexer(CreateEntityId(node.GetTypeName(),DTDEntity::eType));
-                    string ns(m_TargetNamespace);
-                    m_TargetNamespace = node.GetNamespaceName();
-                    ParseContent(node);
-                    m_TargetNamespace = ns;
-                    node.SetTypeIfUnknown(DTDElement::eEmpty);
-// this is not always correct, but it seems that local elements
-// defined by means of global types should be made global as well
-                    if (node.IsNamed() && node.IsEmbedded()) {
-                        node.SetEmbedded(false);
+                    PushEntityLexer(node.GetTypeName());
+                    bool elementForm = m_ElementFormDefault;
+                    bool hasContents = ParseContent(node);
+                    node.SetTypeIfUnknown(hasContents ? DTDElement::eEmpty : DTDElement::eString);
+
+// Make local elements defined by means of global types global.
+// In fact, this is incorrect; also, in case of unqualified form default we must keep
+// such elements embedded, otherwise they will be treated as ns-qualified.
+
+// for us, this trick solves the problem of recursive type definitions:
+// local element A contains local element B, which contains local element A, etc.
+// the way it is now, code generator will simply crash.
+// The better solution would be to modify C++ code generation, of course.
+
+// as of 24may11, the code generator is modified.
+// BUT, the mistake is already made; we want to provide backward compatibility now.
+                    if (node.IsNamed() && node.IsEmbedded() && elementForm) {
+
+                        map<string,DTDElement>::iterator k;
+                        for (k = m_MapElement.begin(); k != m_MapElement.end(); ++k) {
+                            if (!k->second.IsEmbedded() && k->second.IsNamed() &&
+                                k->second.GetName() == node.GetName() && 
+                                k->second.GetTypeName() != node.GetTypeName()) {
+                                break;
+                            }
+                        }
+                        if (k == m_MapElement.end()) {
+                            node.SetEmbedded(false);
+                        }
                     }
                 } else if ( node.GetType() == DTDElement::eUnknownGroup) {
                     found = true;
-                    PushEntityLexer(CreateEntityId(node.GetTypeName(),DTDEntity::eGroup));
+                    PushEntityLexer(node.GetTypeName());
                     ParseGroupRef(node);
                     if (node.GetType() == DTDElement::eUnknownGroup) {
                         node.SetType(DTDElement::eEmpty);
                     }
                 }
-                else if (processed.find(i->second.GetName()) == processed.end()) {
+                else if (processed.find(i->second.GetName() + i->second.GetNamespaceName())
+                            == processed.end()) {
                     if (node.GetType() < DTDElement::eWsdlService) {
-                        PushEntityLexer(CreateEntityId(node.GetTypeName(),DTDEntity::eType));
+                        PushEntityLexer(node.GetTypeName());
                         ParseContent(node);
                     }
                 }
-                processed.insert(i->second.GetName());
+                processed.insert(i->second.GetName() + i->second.GetNamespaceName());
+            }
+        }
+    } while (found);
+
+    do {
+        found = false;
+        map<string,DTDElement>::iterator i;
+        for (i = m_MapElement.begin(); i != m_MapElement.end(); ++i) {
+
+            DTDElement& node = i->second;
+            if (node.HasAttributes()) {
+                list<DTDAttribute>& atts = node.GetNonconstAttributes();
+                list<DTDAttribute>::iterator a;
+                for (a = atts.begin(); a != atts.end(); ++a) {
+
+                    if (!a->GetTypeName().empty()) { 
+                        if ( a->GetType() == DTDAttribute::eUnknown) {
+                            found = true;
+                            PushEntityLexer(a->GetTypeName());
+                            ParseContent(*a);
+                            if (a->GetType() == DTDAttribute::eUnknown) {
+                                a->SetType(DTDAttribute::eString);
+                            }
+                        } else if ( a->GetType() == DTDAttribute::eUnknownGroup) {
+                            found = true;
+                            PushEntityLexer(a->GetTypeName());
+                            atts.erase(a);
+                            ParseAttributeGroupRef(node);
+                            break;
+                        }
+                    }
+                }
             }
         }
     } while (found);
@@ -1387,37 +1495,6 @@ void XSDParser::ProcessNamedTypes(void)
         }
     } while (found);
 
-    do {
-        found = false;
-        map<string,DTDElement>::iterator i;
-        for (i = m_MapElement.begin(); i != m_MapElement.end(); ++i) {
-
-            DTDElement& node = i->second;
-            if (node.HasAttributes()) {
-                list<DTDAttribute>& atts = node.GetNonconstAttributes();
-                list<DTDAttribute>::iterator a;
-                for (a = atts.begin(); a != atts.end(); ++a) {
-
-                    if (!a->GetTypeName().empty()) { 
-                        if ( a->GetType() == DTDAttribute::eUnknown) {
-                            found = true;
-                            PushEntityLexer(CreateEntityId(a->GetTypeName(),DTDEntity::eType));
-                            ParseContent(*a);
-                            if (a->GetType() == DTDAttribute::eUnknown) {
-                                a->SetType(DTDAttribute::eString);
-                            }
-                        } else if ( a->GetType() == DTDAttribute::eUnknownGroup) {
-                            found = true;
-                            PushEntityLexer(CreateEntityId(a->GetTypeName(),DTDEntity::eAttGroup));
-                            atts.erase(a);
-                            ParseAttributeGroupRef(node);
-                            break;
-                        }
-                    }
-                }
-            }
-        }
-    } while (found);
     {
         map<string,DTDElement>::iterator i;
         for (i = m_MapElement.begin(); i != m_MapElement.end(); ++i) {
@@ -1427,12 +1504,18 @@ void XSDParser::ProcessNamedTypes(void)
     m_ResolveTypes = false;
 }
 
-void XSDParser::BeginScope(void)
+void XSDParser::BeginScope(DTDEntity* ent)
 {
     m_StackPrefixToNamespace.push(m_PrefixToNamespace);
     m_StackNamespaceToPrefix.push(m_NamespaceToPrefix);
     m_StackTargetNamespace.push(m_TargetNamespace);
     m_StackElementFormDefault.push(m_ElementFormDefault);
+    m_StackAttributeFormDefault.push(m_AttributeFormDefault);
+    if (ent && ent->GetType() != DTDEntity::eEntity) {
+        ent->GetParseAttributes(m_TargetNamespace,
+            m_ElementFormDefault,m_AttributeFormDefault,
+            m_PrefixToNamespace);
+    }
 }
 void XSDParser::EndScope(void)
 {
@@ -1440,17 +1523,20 @@ void XSDParser::EndScope(void)
     m_NamespaceToPrefix = m_StackNamespaceToPrefix.top();
     m_TargetNamespace = m_StackTargetNamespace.top();
     m_ElementFormDefault = m_StackElementFormDefault.top();
+    m_AttributeFormDefault = m_StackAttributeFormDefault.top();
 
     m_StackPrefixToNamespace.pop();
     m_StackNamespaceToPrefix.pop();
     m_StackTargetNamespace.pop();
     m_StackElementFormDefault.pop();
+    m_StackAttributeFormDefault.pop();
 }
 
-void XSDParser::PushEntityLexer(const string& name)
+DTDEntity* XSDParser::PushEntityLexer(const string& name)
 {
-    DTDParser::PushEntityLexer(name);
-    BeginScope();
+    DTDEntity* ent = DTDParser::PushEntityLexer(name);
+    BeginScope(ent);
+    return ent;
 }
 
 bool XSDParser::PopEntityLexer(void)
@@ -1482,6 +1568,8 @@ void XSDParser::PrintDocumentTree(void)
     
     cout << " === Element form default ===" << endl;
     cout << (m_ElementFormDefault ? "qualified" : "unqualified") << endl;
+    cout << " === Attribute form default ===" << endl;
+    cout << (m_AttributeFormDefault ? "qualified" : "unqualified") << endl;
     cout << endl;
 
     DTDParser::PrintDocumentTree();
